import asyncio
from collections import deque
import copy
import logging
import traceback
import typing
import uuid

import aio_pika

from . import defaults
from . import utils

_LOGGER = logging.getLogger(__name__)


class BroadcastMessage:
    BODY = 'body'
    SENDER = 'sender'
    SUBJECT = 'subject'
    CORRELATION_ID = 'correlation_id'

    @staticmethod
    def create(body, sender=None, subject=None, correlation_id=None):
        message_dict = {
            BroadcastMessage.BODY: body,
            BroadcastMessage.SENDER: sender,
            BroadcastMessage.SUBJECT: subject,
            BroadcastMessage.CORRELATION_ID: correlation_id,
        }
        return message_dict


class BaseConnectionWithExchange:
    """
    An RMQ connection with a channel and exchange
    """
    DEFAULT_EXCHANGE_PARAMS = {'type': aio_pika.ExchangeType.TOPIC}

    def __init__(self, connection, exchange_name=defaults.MESSAGE_EXCHANGE, exchange_params=None, testing_mode=False):
        """
        :type connection: :class:`aio_pika.Connection`
        :type exchange_name: str
        :type exchange_params: dict or NoneType
        """
        super().__init__()

        if exchange_params is None:
            exchange_params = copy.copy(self.DEFAULT_EXCHANGE_PARAMS)

        if testing_mode:
            exchange_params.setdefault('auto_delete', testing_mode)

        self._connection = connection
        self._exchange_name = exchange_name
        self._exchange_params = exchange_params
        self._loop = self._connection.loop

        self._channel = None  # type: typing.Optional[aio_pika.Channel]
        self._exchange = None  # type: typing.Optional[aio_pika.Exchange]
        self._is_closing = False

    @property
    def is_closing(self):
        return self._is_closing

    def loop(self) -> asyncio.BaseEventLoop:
        return self._loop

    def get_exchange_name(self):
        return self._exchange_name

    def channel(self):
        return self._channel

    async def connect(self):
        if self._channel:
            return

        # Create the channel
        self._channel = await self._connection.channel()
        # Create the exchange
        self._exchange = await self._channel.declare_exchange(name=self.get_exchange_name(), **self._exchange_params)

    async def disconnect(self):
        if not self.is_closing:
            self._is_closing = True
            if self.channel() is not None:
                await self.channel().close()
                self._channel = None


class BasePublisherWithReplyQueue:
    """
    A base class for any object that needs to be able to publish a message and to potentially expect a reply.
    """
    # pylint: disable=too-many-instance-attributes

    DEFAULT_EXCHANGE_PARAMS = {'type': aio_pika.ExchangeType.TOPIC}

    def __init__(self,
                 connection,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 exchange_params=None,
                 encoder=defaults.ENCODER,
                 decoder=defaults.DECODER,
                 confirm_deliveries=True,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        """
        :param connection: The aio_pika RMQ connection
        :type connection: :class:`aio_pika.connection.Connection`
        :param exchange_name:
        :param exchange_params:
        :param encoder:
        :param decoder:
        :param confirm_deliveries:
        """
        super().__init__()

        if exchange_params is None:
            exchange_params = copy.copy(self.DEFAULT_EXCHANGE_PARAMS)

        if testing_mode:
            exchange_params.setdefault('auto_delete', testing_mode)

        self._exchange_name = exchange_name
        self._exchange_params = exchange_params
        self._encode = encoder
        self._response_decode = decoder
        self._confirm_deliveries = confirm_deliveries
        if self._confirm_deliveries:
            self._num_published = 0
            self._delivery_info = deque()
        self._testing_mode = testing_mode

        self._awaiting_response = {}

        self._connection = connection
        self._channel = None  # type: typing.Optional[aio_pika.Channel]
        self._exchange = None  # type: typing.Optional[aio_pika.Exchange]
        self._reply_queue = None  # type: typing.Optional[aio_pika.Queue]

        self._is_closing = False

    @property
    def is_closing(self):
        return self._is_closing

    @property
    def is_connected(self):
        return self._channel

    async def connect(self):
        if self.is_connected:
            return

        self._channel = await self._connection.channel(publisher_confirms=self._confirm_deliveries,
                                                       on_return_raises=True)
        self._channel.add_close_callback(self._on_channel_close)

        self._exchange = await self._channel.declare_exchange(name=self.get_exchange_name(), **self._exchange_params)

        # Declare the reply queue
        reply_queue_name = "{}-reply-{}".format(self._exchange_name, str(uuid.uuid4()))
        self._reply_queue = await self._channel.declare_queue(name=reply_queue_name,
                                                              exclusive=True,
                                                              auto_delete=self._testing_mode,
                                                              arguments={"x-expires": defaults.REPLY_QUEUE_EXPIRES})

        await self._reply_queue.bind(self._exchange, routing_key=reply_queue_name)
        await self._reply_queue.consume(self._on_response, no_ack=True)

    async def disconnect(self):
        if not self.is_closing:
            self._is_closing = True
            if self.channel() is not None and not self.channel().is_closed:
                await self._channel.close()
            self._channel = None

    def action_message(self, message):
        """
        Execute a message that involves communication.  This could mean that the
        message gets queued first and then sent as soon as the connection is open.
        In any case the method returns a future for the message.

        :param message: The message to execute
        :return: A future corresponding to action
        :rtype: :class:`kiwipy.Future`
        """
        message.send(self)
        return message.future

    async def publish(self, message, routing_key, mandatory=True):
        """
        Send a fire-and-forget message i.e. no response expected.

        :param message: The message to send
        :param routing_key: The routing key
        :param mandatory: If the message cannot be routed this will raise an UnroutableException
        :return:
        """
        result = await self._exchange.publish(message, routing_key=routing_key, mandatory=mandatory)
        return result

    async def publish_expect_response(self, message, routing_key, mandatory=True):
        # If there is no correlation id we have to set on so that we know what the response will be to
        if not message.correlation_id:
            message.correlation_id = str(uuid.uuid4())
        correlation_id = message.correlation_id

        response_future = asyncio.Future()
        self._awaiting_response[correlation_id] = response_future
        result = await self.publish(message, routing_key=routing_key, mandatory=mandatory)
        return result, response_future

    def get_exchange_name(self):
        return self._exchange_name

    def channel(self):
        return self._channel

    # region RMQ communications

    async def _on_response(self, message):
        """
        Called when we get a message on our response queue

        :param message: The response message
        :type message: :class:`aio_pika.message.IncomingMessage`
        """
        correlation_id = message.correlation_id
        try:
            response_future = self._awaiting_response.pop(correlation_id)
        except KeyError:
            _LOGGER.error("Got a response for an unknown id '%s':\n%s", correlation_id, message)
        else:
            try:
                response = self._response_decode(message.body)
            except Exception:
                _LOGGER.error("Failed to decode message body:\n%s%s", message.body, traceback.format_exc())
                raise
            else:
                utils.response_to_future(response, response_future)
                try:
                    # If the response was a future it means we should get another message that
                    # resolves that future
                    if asyncio.isfuture(response_future.result()):
                        self._awaiting_response[correlation_id] = response_future.result()
                except Exception:  # pylint: disable=broad-except
                    pass

    def _on_channel_close(self, _closing_future):
        """ Reset all channel specific members """
        if self._confirm_deliveries:
            self._num_published = 0
            self._delivery_info = deque()

    # endregion



import concurrent.futures

__all__ = ('RemoteException', 'DeliveryFailed', 'TaskRejected', 'UnroutableError', 'TimeoutError',
           'DuplicateSubscriberIdentifier', 'CommunicatorClosed', 'InvalidStateError', 'QueueEmpty')


class RemoteException(Exception):
    """An exception occurred at the remote end of the call """


class DeliveryFailed(Exception):
    """Failed to deliver a message """


class UnroutableError(DeliveryFailed):
    """The messages was unroutable """


class TaskRejected(Exception):
    """ A task was rejected at the remote end """


class QueueEmpty(Exception):
    """Could not get the next message from the queue because it is empty"""


class DuplicateSubscriberIdentifier(Exception):
    """Failed to add a subscriber because the identifier supplied is already in use"""


class CommunicatorClosed(Exception):
    """Raised when an operation is attempted on a closed communicator"""


class InvalidStateError(Exception):
    """Raise when an object is in an invalid state to perform the desired action"""


TimeoutError = concurrent.futures.TimeoutError  # pylint: disable=redefined-builtin



from .communicator import *
from .tasks import *
from .threadcomms import *
from .utils import *

# pylint: disable=undefined-variable
__all__ = (tasks.__all__ + communicator.__all__ + threadcomms.__all__ + utils.__all__)



__all__ = ('connect',)

DEFAULT_COMM_URI = 'amqp://guest:guest@127.0.0.1/'


def connect(uri: str = DEFAULT_COMM_URI, **kwargs):
    """Create a connection using a URI"""
    if uri.startswith('amqp'):
        from . import rmq  # Avoid circular ref: pylint: disable=import-outside-toplevel
        return rmq.connect(connection_params=uri, **kwargs)

    raise ValueError("Uknown communicator uri '{}'".format(uri))



import asyncio
import collections
import contextlib
import logging
import uuid
from typing import Optional
import weakref

import aio_pika
from async_generator import async_generator, yield_, asynccontextmanager
import shortuuid

import kiwipy
from . import defaults
from . import messages
from . import utils

_LOGGER = logging.getLogger(__name__)

__all__ = 'RmqTaskSubscriber', 'RmqTaskPublisher', 'RmqTaskQueue', 'RmqIncomingTask'

TaskInfo = collections.namedtuple('TaskBody', ('task', 'no_reply'))


class RmqTaskSubscriber(messages.BaseConnectionWithExchange):
    """
    Listens for tasks coming in on the RMQ task queue
    """
    TASK_QUEUE_ARGUMENTS = {"x-message-ttl": defaults.TASK_MESSAGE_TTL}

    def __init__(self,
                 connection: aio_pika.Connection,
                 exchange_name: str = defaults.MESSAGE_EXCHANGE,
                 queue_name: str = defaults.TASK_QUEUE,
                 testing_mode=False,
                 decoder=defaults.DECODER,
                 encoder=defaults.ENCODER,
                 exchange_params=None,
                 prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 prefetch_count=defaults.TASK_PREFETCH_COUNT):
        # pylint: disable=too-many-arguments
        """
        :param connection: An RMQ connection
        :param exchange_name: the name of the exchange to use
        :param queue_name: the name of the task queue to use
        :param decoder: A message decoder
        :param encoder: A response encoder
        """
        super().__init__(connection,
                         exchange_name=exchange_name,
                         exchange_params=exchange_params,
                         testing_mode=testing_mode)

        self._task_queue_name = queue_name
        self._testing_mode = testing_mode
        self._decode = decoder
        self._encode = encoder
        self._prefetch_size = prefetch_size
        self._prefetch_count = prefetch_count
        self._consumer_tag = None

        self._task_queue = None  # type: Optional[aio_pika.Queue]
        self._subscribers = {}
        self._pending_tasks = []

    async def add_task_subscriber(self, subscriber, identifier=None):
        identifier = identifier or shortuuid.uuid()
        if identifier in self._subscribers:
            raise kiwipy.DuplicateSubscriberIdentifier("Task identifier '{}'".format(identifier))

        self._subscribers[identifier] = subscriber
        if self._consumer_tag is None:
            self._consumer_tag = await self._task_queue.consume(self._on_task)

        return identifier

    async def remove_task_subscriber(self, identifier):
        try:
            self._subscribers.pop(identifier)
        except KeyError:
            raise ValueError("Unknown task subscriber '{}'".format(identifier))
        if not self._subscribers:
            await self._task_queue.cancel(self._consumer_tag)
            self._consumer_tag = None

    async def connect(self):
        if self.channel():
            # Already connected
            return

        await super().connect()
        await self.channel().set_qos(prefetch_count=self._prefetch_count, prefetch_size=self._prefetch_size)

        await self._create_task_queue()

    @async_generator
    async def __aiter__(self):
        tasks = []
        try:
            while True:
                task = RmqIncomingTask(self, await self._task_queue.get(timeout=1.))
                tasks.append(task)
                await yield_(task)
        except aio_pika.exceptions.QueueEmpty:
            return
        finally:
            # Put back any tasks that are still pending (i.e. not processed or to be processed)
            for task in tasks:
                if task.state == TASK_PENDING:
                    task.requeue()

    @asynccontextmanager
    @async_generator
    async def next_task(self, no_ack=False, fail=True, timeout=defaults.TASK_FETCH_TIMEOUT):
        """
        Get the next task from the queue.

        raises:
            kiwipy.exceptions.QueueEmpty: When the queue has no tasks within the timeout
        """
        try:
            message = await self._task_queue.get(no_ack=no_ack, fail=fail, timeout=timeout)
        except aio_pika.exceptions.QueueEmpty as exc:
            raise kiwipy.exceptions.QueueEmpty(str(exc))
        else:
            task = RmqIncomingTask(self, message)
            try:
                await yield_(task)
            finally:
                if task.state == TASK_PENDING:
                    task.requeue()

    async def _create_task_queue(self):
        """Create and bind the task queue"""
        arguments = dict(self.TASK_QUEUE_ARGUMENTS)
        if self._testing_mode:
            arguments['x-expires'] = defaults.TEST_QUEUE_EXPIRES

        # x-expires means how long does the queue stay alive after no clients
        # x-message-ttl means what is the default ttl for a message arriving in the queue
        self._task_queue = await self._channel.declare_queue(name=self._task_queue_name,
                                                             durable=not self._testing_mode,
                                                             arguments=arguments)
        await self._task_queue.bind(self._exchange, routing_key=self._task_queue.name)

    async def _on_task(self, message: aio_pika.IncomingMessage):
        """
        :param message: The aio_pika RMQ message
        """
        # Decode the message tuple into a task body for easier use
        rmq_task = RmqIncomingTask(self, message)
        with rmq_task.processing() as outcome:
            for subscriber in self._subscribers.values():
                try:
                    subscriber = utils.ensure_coroutine(subscriber)
                    result = await subscriber(self, rmq_task.body)

                    # If a task returns a future it is not considered done until the chain of
                    # futures (i.e. if the first future resolves to a future and so on) finishes
                    # and produces a concrete result
                    while asyncio.isfuture(result):
                        if not rmq_task.no_reply:
                            await self._send_response(utils.pending_response(), message)
                        result = await result
                except kiwipy.TaskRejected:
                    # Task was rejected by this subscriber, keep trying
                    continue
                except kiwipy.CancelledError:
                    # The subscriber has cancelled their processing of the task
                    outcome.cancel()
                except Exception as exc:  # pylint: disable=broad-except
                    # There was an exception during the processing of this task
                    outcome.set_exception(exc)
                    _LOGGER.exception("Exception occurred while processing task.")
                else:
                    # All good
                    outcome.set_result(result)
                    break  # Got handled

    def _build_response_message(self, body, incoming_message):
        """
        Create a aio_pika Message as a response to a task being deal with.

        :param body: The message body dictionary
        :type body: dict
        :param incoming_message: The original message we are responding to
        :type incoming_message: :class:`aio_pika.IncomingMessage`
        :return: The response message
        :rtype: :class:`aio_pika.Message`
        """
        # Add host info
        body[utils.HOST_KEY] = utils.get_host_info()
        message = aio_pika.Message(body=self._encode(body), correlation_id=incoming_message.correlation_id)

        return message

    async def _send_response(self, msg_body, incoming_message):
        msg = self._build_response_message(msg_body, incoming_message)
        await self._exchange.publish(msg, routing_key=incoming_message.reply_to)


TASK_PENDING = 'pending'
TASK_FINISHED = 'finished'
TASK_PROCESSING = 'processing'
TASK_REQUEUED = 'requeued'


class RmqIncomingTask:

    def __init__(self, subscriber: RmqTaskSubscriber, message: aio_pika.IncomingMessage):
        self._subscriber = subscriber
        self._message = message
        self._task_info = TaskInfo(*subscriber._decode(message.body))
        self._state = TASK_PENDING
        self._outcome_ref = None  # type: Optional[weakref.ReferenceType]

    @property
    def body(self) -> str:
        return self._task_info.task

    @property
    def no_reply(self) -> bool:
        return self._task_info.no_reply

    @property
    def state(self) -> str:
        return self._state

    def process(self) -> asyncio.Future:
        if self._state != TASK_PENDING:
            raise asyncio.InvalidStateError("The task is {}".format(self._state))

        self._state = TASK_PROCESSING
        outcome = self._create_future()
        # Rely on the done callback to signal the end of processing
        outcome.add_done_callback(self._task_done)
        # Or the user let's the future get destroyed
        self._outcome_ref = weakref.ref(outcome, self._outcome_destroyed)

        return outcome

    def requeue(self):
        if self._state not in [TASK_PENDING, TASK_PROCESSING]:
            raise asyncio.InvalidStateError("The task is {}".format(self._state))

        self._state = TASK_REQUEUED
        self._message.nack(requeue=True)
        self._finalise()

    @contextlib.contextmanager
    def processing(self):
        """Processing context.  The task should be done at the end otherwise it's assumed the
        caller doesn't want to process it and it's sent back to the queue"""
        if self._state != TASK_PENDING:
            raise asyncio.InvalidStateError("The task is {}".format(self._state))

        self._state = TASK_PROCESSING
        outcome = self._subscriber.loop().create_future()
        try:
            yield outcome
        except KeyboardInterrupt:  # pylint: disable=try-except-raise
            raise
        except Exception as exc:
            # Set the exception on the task and re-raise so the client also sees it
            outcome.set_exception(exc)
            raise
        finally:
            if outcome.done():
                self._task_done(outcome)
            else:
                self.requeue()

    def _task_done(self, outcome: asyncio.Future):
        assert outcome.done()
        self._outcome_ref = None

        if outcome.cancelled():
            # Whoever took the task decided not to process it
            self._state = TASK_PENDING
        else:
            # Permanently store the outcome
            self._state = TASK_FINISHED
            self._message.ack()
            if not self.no_reply:
                # Schedule a task to send the appropriate response
                if outcome.exception():
                    reply_body = utils.exception_response(outcome.exception())
                else:
                    reply_body = utils.result_response(outcome.result())
                # pylint: disable=protected-access
                self._subscriber.loop().create_task(self._subscriber._send_response(reply_body, self._message))

        # Clean up
        self._finalise()

    def _outcome_destroyed(self, outcome_ref):
        # This only happens if someone called self.process() and then let the future
        # get destroyed without setting an outcome
        assert outcome_ref is self._outcome_ref
        # This task will not be processed
        self._outcome_ref = None
        self.requeue()

    def _finalise(self):
        self._outcome_ref = None
        self._subscriber = None
        self._message = None

    def _create_future(self) -> asyncio.Future:
        return self._subscriber.loop().create_future()


class RmqTaskPublisher(messages.BasePublisherWithReplyQueue):
    """
    Publishes messages to the RMQ task queue and gets the response
    """

    def __init__(self,
                 connection,
                 queue_name=defaults.TASK_QUEUE,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 exchange_params=None,
                 encoder=defaults.ENCODER,
                 decoder=defaults.DECODER,
                 confirm_deliveries=True,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        super().__init__(connection,
                         exchange_name=exchange_name,
                         exchange_params=exchange_params,
                         encoder=encoder,
                         decoder=decoder,
                         confirm_deliveries=confirm_deliveries,
                         testing_mode=testing_mode)
        self._task_queue_name = queue_name

    async def task_send(self, task, no_reply: bool = False) -> asyncio.Future:
        """Send a task for processing by a task subscriber.

        All task messages will be set to be persistent by setting `delivery_mode=2`.

        :param task: The task payload
        :param no_reply: Don't send a reply containing the result of the task
        :return: A future representing the result of the task
        """
        # Build the full message body and encode as a tuple
        body = self._encode((task, no_reply))
        # Now build up the full aio_pika message
        task_msg = aio_pika.Message(
            body=body,
            correlation_id=str(uuid.uuid4()),
            reply_to=self._reply_queue.name,
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT  # Task messages need to be persistent
        )

        result_future = None
        if no_reply:
            published = await self.publish(task_msg, routing_key=self._task_queue_name, mandatory=True)
        else:
            published, result_future = await self.publish_expect_response(task_msg,
                                                                          routing_key=self._task_queue_name,
                                                                          mandatory=True)

        assert published, "The task was not published to the exchange"
        return result_future


class RmqTaskQueue:
    """Combines a task publisher and subscriber to create a work queue where you can do both"""

    def __init__(self,
                 connection,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 queue_name=defaults.TASK_QUEUE,
                 decoder=defaults.DECODER,
                 encoder=defaults.ENCODER,
                 exchange_params=None,
                 prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 prefetch_count=defaults.TASK_PREFETCH_COUNT,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        self._publisher = RmqTaskPublisher(connection,
                                           exchange_name=exchange_name,
                                           exchange_params=exchange_params,
                                           queue_name=queue_name,
                                           decoder=decoder,
                                           encoder=encoder,
                                           testing_mode=testing_mode)
        self._subscriber = RmqTaskSubscriber(connection,
                                             exchange_name=exchange_name,
                                             exchange_params=exchange_params,
                                             queue_name=queue_name,
                                             decoder=decoder,
                                             encoder=encoder,
                                             prefetch_size=prefetch_size,
                                             prefetch_count=prefetch_count,
                                             testing_mode=testing_mode)

    @async_generator
    async def __aiter__(self):
        # Have to do it this way rather than the more convenient yield from style because
        # python doesn't support it for coroutines.  See:
        # https://stackoverflow.com/questions/47376408/why-cant-i-yield-from-inside-an-async-function
        async for task in self._subscriber:
            await yield_(task)

    async def task_send(self, task, no_reply: bool = False):
        """Send a task to the queue"""
        return await self._publisher.task_send(task, no_reply)

    async def add_task_subscriber(self, subscriber, identifier=None):
        return await self._subscriber.add_task_subscriber(subscriber, identifier)

    async def remove_task_subscriber(self, identifier):
        return await self._subscriber.remove_task_subscriber(identifier)

    @asynccontextmanager
    @async_generator
    async def next_task(self, no_ack=False, fail=True, timeout=defaults.TASK_FETCH_TIMEOUT):
        async with self._subscriber.next_task(no_ack=no_ack, fail=fail, timeout=timeout) as task:
            await yield_(task)

    async def connect(self):
        await self._subscriber.connect()
        await self._publisher.connect()

    async def disconnect(self):
        await self._subscriber.disconnect()
        await self._publisher.disconnect()



from functools import partial
import yaml

# Times are in milliseconds
SECONDS_TO_MILLISECONDS = 1000

TASK_EXCHANGE = 'kiwipy.tasks'
TASK_QUEUE = 'kiwipy.tasks'
MESSAGE_EXCHANGE = 'kiwipy.messages'
BROADCAST_TOPIC = '[broadcast]'
RPC_TOPIC = '[rpc]'
# Have to set MESSAGE_TTL to > 65535 because of a bug in aio-pika which fails when using RabbitMQ
# 3.5 (as present in Ubuntu 16.04) and likely earlier.  If you have a newer version of RabbitMQ
# it's fine to set this lower.  See:
# https://github.com/mosquito/aio-pika/issues/165
MESSAGE_TTL = 66 * SECONDS_TO_MILLISECONDS
TEST_QUEUE_EXPIRES = 10 * SECONDS_TO_MILLISECONDS
QUEUE_EXPIRES = 60 * SECONDS_TO_MILLISECONDS
REPLY_QUEUE_EXPIRES = 60 * SECONDS_TO_MILLISECONDS
TASK_MESSAGE_TTL = 60 * SECONDS_TO_MILLISECONDS * 60 * 24 * 7 * 120  # 120 days
TASK_PREFETCH_SIZE = 0
TASK_PREFETCH_COUNT = 0
TASK_FETCH_TIMEOUT = 5.

ENCODER = partial(yaml.dump, encoding='utf-8')
DECODER = partial(yaml.load, Loader=yaml.FullLoader)



from typing import Callable, Any

import kiwipy  # pylint: disable=unused-import

# RPC subscriber params: communicator, msg
RpcSubscriber = Callable[['kiwipy.Communicator', Any], Any]
# Task subscriber params: communicator, task
TaskSubscriber = Callable[['kiwipy.Communicator', Any], Any]
# Broadcast subscribers params: communicator, body, sender, subject, correlation id
BroadcastSubscriber = Callable[['kiwipy.Communicator', Any, Any, Any, Any], Any]



import concurrent.futures
import contextlib
import logging

__all__ = 'Future', 'chain', 'copy_future', 'CancelledError', 'capture_exceptions', 'wait', 'as_completed'

_LOGGER = logging.getLogger(__name__)

CancelledError = concurrent.futures.CancelledError
wait = concurrent.futures.wait  # pylint: disable=invalid-name
as_completed = concurrent.futures.as_completed  # pylint: disable=invalid-name
Future = concurrent.futures.Future


def copy_future(source, target):
    """ Copy the status of future a to b unless b is already done in
    which case return

    :param source: The source future
    :type source: :class:`Future`
    :param target: The target future
    :type target: :class:`Future`
    """
    if target.done():
        return

    if source.cancelled():
        target.cancel()
    else:
        with capture_exceptions(target):
            target.set_result(source.result())


def chain(source, target):
    """Chain two futures together so that when one completes, so does the other.

    The result (success or failure) of ``a`` will be copied to ``b``, unless
    ``b`` has already been completed or cancelled by the time ``a`` finishes.
    """
    source.add_done_callback(lambda first: copy_future(first, target))


@contextlib.contextmanager
def capture_exceptions(future, ignore=()):
    """
    Capture any exceptions in the context and set them as the result of the given future

    :param future: The future to the exception on
    :type future: :class:`kiwipy.Future`
    :param ignore: An optional list of exception types to ignore, these will be raised and not set on the future
    """
    try:
        yield
    except Exception as exception:  # pylint: disable=broad-except
        if isinstance(exception, ignore):
            raise

        future.set_exception(exception)
